"""
    flaskext.slithers
    ~~~~~~~~~~~~~~~~~

    Flask extension for quickly building RESTful API. Based on flask-slither,
    but is now somewhat more database agnostic and has an improved workflow.

    :copyright: (c) 2015 by Nico Gevers.
    :license: MIT, see LICENSE for more details.
"""
import inflect

__author__ = 'Nico Gevers'
__version__ = (1, 1, 6)


def register_resource(mod, view, **kwargs):
    """Register the resource on the resource name or a custom url"""
    resource_name = view.__name__.lower()[:-8]
    endpoint = kwargs.get('endpoint', "{}_api".format(resource_name))
    plural_resource_name = inflect.engine().plural(resource_name)
    path = kwargs.get('url', plural_resource_name).strip('/')
    url = '/{}'.format(path)
    setattr(view, '_url', url)  # need this for 201 location header
    view_func = view.as_view(endpoint)

    mod.add_url_rule(url, view_func=view_func,
                     methods=['GET', 'POST', 'OPTIONS'])
    mod.add_url_rule('{}/<obj_id>'.format(url),
                     view_func=view_func,
                     methods=['GET', 'PATCH', 'PUT', 'DELETE', 'OPTIONS'])


class ApiException(Exception):
    def __init__(self, message, status):
        self.message = message
        self.status = status
        Exception.__init__(
            self, "API Exception: {}, {}".format(status, message))



# -*- coding: utf-8 -*-
from bson.objectid import ObjectId
from bson.errors import InvalidId
from pymongo import MongoClient
from uuid import UUID
from datetime import datetime

import json
import logging


class JSONEncoder(json.JSONEncoder):
    """Encode all fancy mongo objects into normal strings. While it is useful
       in some cases to use the mongo json_util class, generally we want to
       hide complexity for the API client, and so don't return types. It is
       up to the API to convert incoming types, and up to this class to
       serialize all complex objects to standard JSON types"""

    def default(self, obj):
        if isinstance(obj, ObjectId):
            return str(obj)
        if isinstance(obj, UUID):
            return str(obj).replace('-', '')
        if isinstance(obj, datetime):
            return int(obj.timestamp())
        return json.JSONEncoder.default(self, obj)


class MongoDbQuery():
    """This class encapsulates some method for querying the mongo database."""

    def __init__(self, **kwargs):
        self.collection = kwargs.get('collection', '')
        self.client = MongoClient(
            kwargs.get('DB_HOST', 'localhost'),
            kwargs.get('DB_PORT', 27017))
        db_name = kwargs.get('DB_NAME', 'testing_slither')
        self.db = self.client[db_name]

    def __exit__(self):
        self.db.close()

    def _clean_record(self, record):
        """Remove all fields with `None` values"""
        for k, v in dict(record).items():
            if isinstance(v, dict):
                v = self._clean_record(v)
            if v is None:
                record.pop(k)
        return record

    def get_instance(self, collection, obj_id, **kwargs):
        """Get a record from the database with the id field matching `obj_id`.
        """
        logging.info("Getting single record")
        try:
            obj_id = ObjectId(obj_id)
        except InvalidId:
            logging.error("Invalid ObjectId: {}".format(obj_id))
            return {}
        projection = kwargs.get('projection', {})
        projection = None if len(projection) < 1 else projection
        query = kwargs.get('query', {})
        query.update({'_id': obj_id})
        logging.debug("Query: {}".format(query))
        logging.debug("Projection: {}".format(projection))
        record = self.db[collection].find_one(query, projection)
        return record

    def get_collection(self, collection, **kwargs):
        """Get a record from the database with the id field matching `obj_id`.
        """
        query = kwargs.get('query', {})
        projection = kwargs.get('projection', {})
        projection = None if len(projection) < 1 else projection
        limit = kwargs.get('limit', 0)  # no limit by default
        # sort = kwargs.get('sort', {})
        logging.info("About to get a collection from the database")
        logging.debug("Collection: {}".format(collection))
        logging.debug("Query: {}".format(query))
        logging.debug("Projection: {}".format(projection))
        logging.debug("Limit: {}".format(limit))
        records = \
            list(self.db[collection].find(query, projection).limit(limit))
        logging.debug("Got {} results".format(len(records)))
        return records

    def delete(self, collection, record):
        if record is not None and '_id' in record:
            logging.info("Deleting record: {}".format(record['_id']))
            self.db[collection].remove({'_id': record['_id']})

    def create(self, collection, record):
        logging.info("Creating new record")
        return self.db[collection].insert(self._clean_record(record))

    def update(self, collection, record, orig_record, full_update=False):
        logging.info("Updating record.")
        logging.debug("Full update? {}".format(full_update))
        if '_id' not in record:
            logging.warning("No id in record. Cannot update")
            logging.debug("Record: {}".format(record))
            return
        if len(record) < 1:
            logging.warning("Not updating empty record")
            return
        query = self._clean_record(record) if full_update else {'$set': record}
        record.pop('_id', '')
        _id = orig_record['_id']
        logging.debug("_id: {}".format(_id))
        logging.debug("Query: {}".format(query))
        self.db[collection].update({'_id': _id}, query)
        record['_id'] = _id
        return record

    def serialize(self, root, records):
        """Serialize the payload into JSON"""
        logging.info("Serializing record")
        logging.debug("Root: {}".format(root))
        logging.debug("Records: {}".format(records))
        if records == {}:
            return '{}'
        if isinstance(records, dict):
            if list(records.keys())[0] == 'errors':
                logging.warning("Found errors. Moving on".format(records))
                root = None
            elif '_id' in records:
                records['id'] = records.pop('_id')
        else:
            records = list(records)

            # rename _id to id
            for r in records:
                if '_id' in r:
                    r['id'] = r.pop('_id')

        if root is not None:
            records = {root: records}
        return json.dumps(records, cls=JSONEncoder)



# -*- coding: utf-8 -*-
from flask import make_response, request, g, current_app, json, abort
from flask.views import MethodView
from flask_slither.decorators import endpoint, crossdomain
from flask_slither.db import MongoDbQuery

import time


class BaseResource(MethodView):

    #: A list of HTTP methods that are open for use. Any method not on this
    #: list will return a 405 if accessed."""
    allowed_methods = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']

    #: This is the database table/collection name that this resource will
    #: access by default
    db_collection = None

    #: This class is used for all database queries as well as serialization
    #: of the final records
    db_query = MongoDbQuery

    #: By default request bodies must have the db_collection name as the root
    #: field. Set this to false if not the case
    enforce_json_root = True

    #: Set the json_root manually if it is something other than db_collection
    # json_root = self.db_collection

    #: POST/PUT/PATCH requests must have the root of the data equal to the
    #: `db_collection`. If this is false, it is not enforced
    enforce_payload_collection = True

    #: If an authentication class is defined, it can contain the methods
    #: `is_authenticated` and `is_authorized` which will contain the logic
    #: to check if the request has access to the resource. If the methods
    #: aren't defined, it is assumed that authentication/authorization is
    #: successful.
    authentication = None

    #: If a validation class is defined, the `validate_<type>` method will be
    #: used to validate the type of request (e.g. GET, POST etc). Validation
    #: is assumed to pass if no method is defined.
    validation = None

    #: Allow CORS requests, and if True, put in extra parameters
    cors_enabled = False
    cors_config = {
        'max_age': 21600,
        'methods': ["GET", "POST", "PUT", "PATCH", "DELETE"],
        'allowed': None,
        'blacklist': [],
        'headers': None
    }

    def __init__(self, app=None, **kwargs):
        if app is not None:
            self.app = app
            self.init_app(self.app)
        else:
            self.app = None
        self.db_query = self.db_query(**current_app.config)

    def init_app(self, app):
        app.teardown_appcontext(self.teardown)

    def _exception_handler(self, e):
        """This exception handler catches should only be invoked when we need
           to quit the workflow prematurely. It takes in an `ApiException`
           which will contain the error, details of the error and  status code.
           When the application is in testing mode, it will also return the
           stack trace"""
        if isinstance(e.message, dict):
            return self._prep_response(
                e.message['msg'], status=e.message['status'])
        if e.message.find('Validation') == 0:
            return self._prep_response(self.validation.errors, status=400)
        elif e.message.find('No record') == 0:
            return self._prep_response(status=404)
        else:
            return self._prep_response({'message': e.message},
                                       status=409)

    def _get_instance(self, **kwargs):
        """Loads the record specified by the `obj_id` path in the url and
           stores it in g._resource_instance"""
        current_app.logger.info("Getting instance")
        current_app.logger.debug("kwargs: {}".format(kwargs))

        current_app.logger.info(
            "Loading instance: {}".format(kwargs['obj_id']))
        rec = self.db_query.get_instance(self.db_collection, kwargs['obj_id'])
        g._resource_instance = rec
        current_app.logger.debug(
            "g._resource_instance: {}".format(g._resource_instance))
        return rec

    def _payload_root(self):
        """ Returns the expected json root in the payload"""
        if hasattr(self, 'json_root'):
            return self.json_root
        return self.db_collection

    def _make_response(self, status, data=None, **kwargs):
        if kwargs.get('is_file', False):
            current_app.logger.info("Setting response from first parameter")
            response = data
        else:
            current_app.logger.info("Generating response and sending")
            current_app.logger.debug("Status: {}".format(status))
            current_app.logger.debug("Data: {}".format(data))
            has_errors = str(status)[0] in ['4', '5']
            if has_errors:
                current_app.logger.debug("Returning errors payload")
                if isinstance(data, dict):
                    payload = data if 'errors' in list(data.keys() or '') \
                        else {'errors': data}
                else:
                    payload = {'errors': data}
                payload = self.db_query.serialize(None, payload)
            else:
                if kwargs.get('no_serialize', False):
                    payload = data
                else:
                    payload = "" if data is None else \
                        self.db_query.serialize(self._payload_root(), data)
            current_app.logger.debug("Payload: {}".format(payload))
            response = make_response(payload, status)

            current_app.logger.info("Adding response headers")
            response.headers.add('Cache-Control',
                                 'max-age={},must-revalidate'.format(30))
            for h in kwargs.get('headers', []):
                response.headers.add(h[0], h[1])
            if request.method == 'POST' and not has_errors and status == 201:
                location = "{}/{}".format(self._url, data['id'])
                response.headers.add('location', location)
            response.expires = time.time() + 30

            if data is not None:
                response.mimetype = "application/json"
            current_app.logger.debug("Headers: {}".format(response.headers))
        if kwargs.get('abort', False):
            abort(response)
        return response

    def fiddle_id(self, obj_id):
        """In some cases the `obj_id` in the url doesn't exactly match the
           record id. This method allows for the fiddling of the id to match
           the correct record."""
        return obj_id

    def transform_record(self, request_data):
        """A hook for changing the request data into something that should
           be persisted in the database"""
        return request_data

    def transform_payload(self, payload):
        """A hook for the user to transform the payload before it is sent
           back to the client."""
        return payload

    def post_save(self, record):
        """Hook called after a record is saved."""
        pass

    def post_delete(self, record):
        """Hook called after a record is deleted."""
        pass

    def merge_record_data(self, changes, orig_record=None):
        """This method merges PATCH requests with the db record to ensure no
           data is lost. In addition, it is also a hook for other fields to
           be overwritten, to ensure immutable fields aren't changed by a
           request."""
        current_app.logger.info("Merging request data with db record")
        current_app.logger.debug("orig_record: {}".format(orig_record))
        current_app.logger.debug("Changes".format(changes))
        final_record = changes
        if request.method == 'PATCH':
            final_record = dict(orig_record)
            final_record.update(changes)
        elif request.method == 'PUT':
            if '_id' in orig_record:
                final_record['_id'] = orig_record['_id']
        return final_record

    def access_limits(self, **kwargs):
        """This method returns a base query to be used for db queries"""
        return {}

    def limit_fields(self, **kwargs):
        """This method returns the projections for this resource"""
        return {}

    @crossdomain
    @endpoint
    def get(self, **kwargs):
        current_app.logger.info("GETting record(s) from database")
        records = []

        # generate meta information
        params = {'query': self.access_limits(**kwargs), 'projection': {}}
        if '_limit' in request.args:
            try:
                params['limit'] = int(request.args.get('_limit'))
            except ValueError:
                current_app.logger.debug("No record limit override")
                pass
        if '_fields' in request.args:
            params['projection'] = \
                {r: True for r in request.args.get('_fields', '').split(',')}
        params['projection'].update(self.limit_fields(**kwargs))

        if 'obj_id' in kwargs:
            records = self.db_query.get_instance(
                self.db_collection, kwargs['obj_id'], **params)
            if records in [{}, None]:
                return self._make_response(404)
        else:
            records = \
                self.db_query.get_collection(self.db_collection, **params)

        return self._make_response(200, self.transform_payload(records))

    @crossdomain
    @endpoint
    def post(self, **kwargs):
        current_app.logger.info("POSTing record to database")
        current_app.logger.debug(g._saveable_record)
        record_id = self.db_query.create(self.db_collection,
                                         g._saveable_record)
        record = self.db_query.get_instance(self.db_collection, record_id)
        self.post_save(record)
        return self._make_response(201, record)

    @crossdomain
    @endpoint
    def put(self, **kwargs):
        current_app.logger.info("PUTting record to database")
        current_app.logger.debug(g._saveable_record)
        record = self.db_query.update(self.db_collection, g._saveable_record,
                                      orig_record=g._resource_instance,
                                      full_update=True)
        self.post_save(record)
        return self._make_response(204)

    @crossdomain
    @endpoint
    def patch(self, **kwargs):
        current_app.logger.info("PATCHing record to database")
        current_app.logger.debug(g._saveable_record)
        record = self.db_query.update(self.db_collection, g._saveable_record,
                                      orig_record=g._resource_instance)
        self.post_save(record)
        return self._make_response(204)

    @crossdomain
    @endpoint
    def delete(self, **kwargs):
        current_app.logger.info("DELETEing record from database")
        self.db_query.delete(self.db_collection, g._resource_instance)
        self.post_delete(g._resource_instance)
        return self._make_response(204)

    @crossdomain
    def options(self, **kwargs):
        """This method has been implemented as per the CORS spec, however is
        not accessible by default. To included it make `cors_enabled` = True"""
        if self.cors_enabled:
            return self._make_response(200)
        return self._make_response(405, "CORS request rejected")



"""
Flask-Slither
-------------

Slither is a thin library that sits between a mongodb and RESTful
API endpoints.
"""
from setuptools import setup


setup(
    name='Flask-Slither',
    version='1.1.6',
    url='http://github.com/gevious/flask_slither',
    license='MIT',
    author='Nico Gevers',
    author_email='me@gevious.com',
    description='A small library between MongoDB and JSON API endpoints',
    long_description=__doc__,
    packages=['flask_slither'],
    test_suite='nose.collector',
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'Flask==0.10.1',
        'pymongo==2.6',
        'inflect==0.2.5'
    ],
    tests_require=[
        'Flask==0.10.1',
        'inflect==0.2.5',
        'pymongo==2.6',
        'nose==1.3.3'
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



# -*- coding: utf-8 -*-
from flask import make_response, request, current_app, Response, g, json
from functools import wraps
from urllib.parse import urlparse


def crossdomain(f):
    """This decorator sets the rules for the crossdomain request per http
       method. The settings are taken from the actual resource itself, and
       returned as per the CORS spec.

       All CORS requests are rejected if the resource's `allow_methods`
       doesn't include the 'OPTIONS' method. """
    @wraps(f)
    def decorator(self, *args, **kwargs):
        # TODO: if a non-cors request has the origin header, this will fail
        if not self.cors_enabled and 'origin' in request.headers:
            return self._make_response(405, "CORS request rejected")
        resp = f(self, *args, **kwargs)

        h = resp.headers
        current_app.logger.debug("Request Headers: {}".format(request.headers))
        allowed_methods = self.cors_config['methods'] + ["OPTIONS"]
        h['Access-Control-Allow-Methods'] = ", ".join(allowed_methods)
        h['Access-Control-Max-Age'] = self.cors_config.get('max_age', 21600)

        # Request Origin checks
        hostname = urlparse(request.headers['origin']).netloc \
            if 'origin' in request.headers else request.headers['host']
        if hostname in self.cors_config.get('blacklist', []):
            return self._make_response(405, "CORS request blacklisted")
        if self.cors_config.get('allowed', None) is not None and \
                hostname not in self.cors_config.get('allowed', None):
            return self._make_response(405, "CORS request refused")
        if 'origin' in request.headers:
            h['Access-Control-Allow-Origin'] = request.headers['origin']

        # Request header checks
        if 'access-control-request-headers' in request.headers:
            if self.cors_config.get('headers', None) is None:
                allowed_headers = \
                    request.headers.get('access-control-request-headers', "*")
            else:
                allowed_headers = []
                for k in request.headers.get(
                        'access-control-request-headers', []):
                    if k in self.cors_config.get('headers', []):
                        allowed_headers.append(k)
                allowed_headers = " ,".join(allowed_headers)
            h['Access-Control-Allow-Headers'] = allowed_headers

        return resp
    return decorator


def endpoint(f):
    """This decorator marks this method as an endpoint. It is responsible for
       the request workflow and will call each relevant method in turn."""

    def check_authentication(self, **kwargs):
        """If the `authentication` variable is defined and not None, the
           specified method will be run. On True the request will continue
           otherwise it will fail with a 401 authentication error"""
        if getattr(self, 'authentication', None) is None:
            current_app.logger.debug("No authentication method")
            return

        a = self.authentication()
        if not hasattr(a, 'is_authenticated'):
            current_app.logger.debug("No is_authenticated method")
            return

        if not a.is_authenticated(**kwargs):
            msg = getattr(g, 'authentication_error', 'Authentication failed')
            current_app.logger.warning("Authentication failed: {}".format(msg))
            return self._make_response(401, msg, abort=True)
        current_app.logger.debug("Authentication successful")

    def check_authorization(self):
        """If the `authorization` variable is defined and not None, the
           specified method will be run. On True the request will continue
           otherwise it will fail with a 403 authorization error"""
        current_app.logger.info("Checking authentication/authorization")
        auth_class = getattr(self, 'authorization',
                             getattr(self, 'authentication', None))
        if auth_class is None:
            current_app.logger.debug("No authorization class")
            return

        a = auth_class()
        if not hasattr(a, 'is_authorized'):
            current_app.logger.debug("No is_authorized method")
            return

        if not a.is_authorized(record=g._resource_instance):
            current_app.logger.warning("Authorization failed")
            return self._make_response(403, "Authorization failed", abort=True)
        current_app.logger.debug("Authorization successful")

    def validate_request(self, **kwargs):
        """Call the validator class and validate the request_data. This method
           returns True or False. On False, a 400 will be returned with the
           reasons for the validation error. On True, the operation will
           continue."""
        current_app.logger.info(
            "Checking {} validation".format(request.method))
        if getattr(self, 'validation', None) is None:
            current_app.logger.warning("No validation specified")
            return

        v = self.validation()
        method = 'validate_{}'.format(request.method.lower())

        if not hasattr(v, method):
            current_app.logger.warning("No validation method specified")
            return
        errors = getattr(v, method)(**kwargs)
        current_app.logger.debug("Validation errors: {}".format(errors))

        if errors is not None and len(errors) > 0:
            current_app.logger.warning("Validation errors found")
            self._make_response(400, errors, abort=True)

    def load_request_data(self):
        if request.method in ['GET', 'DELETE']:
            return
        current_app.logger.info("Saving json payload in memory")
        # For now we assume JSON. Later in life we can make this more
        # payload agnostic
        try:
            d = request.data.decode('utf-8')
            g._rq_data = {} if d.strip() == "" else json.loads(d)
        except ValueError:
            return self._make_response(400, "Malformed JSON in request body",
                                       abort=True)

        if self.enforce_json_root and g._rq_data != {} and \
                list(g._rq_data.keys()) != [self._payload_root()]:
            msg = "Invalid JSON root in request body"
            current_app.logger.error(msg)
            current_app.logger.debug(
                "Found {}, expecting {}".format(
                    list(g._rq_data.keys()), self._payload_root()))
            return self._make_response(400, msg, abort=True)
        elif self._payload_root() in list(g._rq_data.keys()):
                current_app.logger.debug("Removing JSON root from rq payload")
                g._rq_data = g._rq_data[self._payload_root()]
        current_app.logger.debug("g._rq_data: {}".format(g._rq_data))
        return g._rq_data

    @wraps(f)
    def decorator(self, *args, **kwargs):
        current_app.logger.info("Got {} request".format(request.method))
        current_app.logger.info("Endpoint: {}".format(request.url))
        if request.method not in self.allowed_methods:
            msg = "Request method {} is unavailable".format(request.method)
            current_app.logger.error(msg)
            return self._make_response(405, msg, abort=True)

        current_app.logger.info("Checking db table/collection is defined")
        if self.enforce_payload_collection and self.db_collection is None:
            msg = "No DB collection defined"
            current_app.logger.error(msg)
            return make_response(Response(msg, 424))

        if request.method in ['POST', 'PUT', 'PATCH']:
            load_request_data(self)

        check_authentication(self, **kwargs)
        if kwargs.get('obj_id', False):
            kwargs['obj_id'] = self.fiddle_id(kwargs['obj_id'])
            self._get_instance(**kwargs)
        else:
            g._resource_instance = {}

        if request.method in ['POST', 'PUT', 'PATCH']:
            check_authorization(self)
            r = self.transform_record(g._rq_data.get(self._payload_root(),
                                                     g._rq_data))
            g._saveable_record = dict(self.merge_record_data(
                r, dict(getattr(g, '_resource_instance', r))))
            validate_request(self, data=g._saveable_record)
        else:
            check_authorization(self)
            validate_request(self)
        return f(self, *args, **kwargs)
    return decorator
